import typing as t
from contextlib import asynccontextmanager
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from google.cloud import firestore
from motor.motor_asyncio import AsyncIOMotorClient, AsyncIOMotorDatabase
from pydantic import SecretStr
from redis.asyncio import Redis
from acb.adapters.nosql._base import NosqlBase, NosqlBaseSettings, NosqlCollection
from acb.adapters.nosql.firestore import (
    Nosql as FirestoreNosql,
)
from acb.adapters.nosql.firestore import (
    NosqlSettings as FirestoreSettings,
)
from acb.adapters.nosql.mongodb import (
    Nosql as MongoDBNosql,
)
from acb.adapters.nosql.mongodb import (
    NosqlSettings as MongoDBSettings,
)
from acb.adapters.nosql.redis import Nosql as RedisNosql
from acb.adapters.nosql.redis import NosqlSettings as RedisSettings


class MockNosqlBase(NosqlBase):
    def __init__(self) -> None:
        super().__init__()
        self.config = MagicMock()
        self.logger = MagicMock()
        self._collections = {}
        self._client = None
        self._db = None
        self._transaction = None
        self._data: t.Dict[str, t.Dict[str, t.Dict[str, t.Any]]] = {}

    async def init(self) -> None:
        self.logger.info("Initializing mock NoSQL adapter")

    async def find(
        self, collection: str, filter: t.Dict[str, t.Any], **kwargs: t.Any
    ) -> t.List[t.Dict[str, t.Any]]:
        if collection not in self._data:
            return []

        results = []
        for doc_id, doc in self._data[collection].items():
            if self._matches_filter(doc, filter):
                doc_with_id = doc.copy()
                doc_with_id["_id"] = doc_id
                results.append(doc_with_id)

        limit = kwargs.get("limit")
        if limit is not None:
            results = results[:limit]

        return results

    async def find_one(
        self, collection: str, filter: t.Dict[str, t.Any], **kwargs: t.Any
    ) -> t.Optional[t.Dict[str, t.Any]]:
        results = await self.find(collection, filter, **kwargs, limit=1)
        return next(iter(results), None)

    async def insert_one(
        self, collection: str, document: t.Dict[str, t.Any], **kwargs: t.Any
    ) -> t.Any:
        if collection not in self._data:
            self._data[collection] = {}

        doc_id = document.get("_id", f"id_{len(self._data[collection]) + 1}")

        if "_id" in document:
            document = document.copy()
            del document["_id"]

        self._data[collection][doc_id] = document

        return doc_id

    async def insert_many(
        self, collection: str, documents: t.List[t.Dict[str, t.Any]], **kwargs: t.Any
    ) -> t.List[t.Any]:
        ids = []
        for doc in documents:
            doc_id = await self.insert_one(collection, doc, **kwargs)
            ids.append(doc_id)
        return ids

    async def update_one(
        self,
        collection: str,
        filter: t.Dict[str, t.Any],
        update: t.Dict[str, t.Any],
        **kwargs: t.Any,
    ) -> t.Any:
        doc = await self.find_one(collection, filter)
        if not doc:
            return {"modified_count": 0}

        doc_id = doc["_id"]

        if "$set" in update:
            for key, value in update["$set"].items():
                self._data[collection][doc_id][key] = value
        else:
            for key, value in update.items():
                self._data[collection][doc_id][key] = value

        return {"modified_count": 1}

    async def update_many(
        self,
        collection: str,
        filter: t.Dict[str, t.Any],
        update: t.Dict[str, t.Any],
        **kwargs: t.Any,
    ) -> t.Any:
        docs = await self.find(collection, filter)
        if not docs:
            return {"modified_count": 0}

        modified_count = 0
        for doc in docs:
            doc_id = doc["_id"]

            if "$set" in update:
                for key, value in update["$set"].items():
                    self._data[collection][doc_id][key] = value
            else:
                for key, value in update.items():
                    self._data[collection][doc_id][key] = value

            modified_count += 1

        return {"modified_count": modified_count}

    async def delete_one(
        self, collection: str, filter: t.Dict[str, t.Any], **kwargs: t.Any
    ) -> t.Any:
        doc = await self.find_one(collection, filter)
        if not doc:
            return {"deleted_count": 0}

        doc_id = doc["_id"]
        del self._data[collection][doc_id]

        return {"deleted_count": 1}

    async def delete_many(
        self, collection: str, filter: t.Dict[str, t.Any], **kwargs: t.Any
    ) -> t.Any:
        docs = await self.find(collection, filter)
        if not docs:
            return {"deleted_count": 0}

        deleted_count = 0
        for doc in docs:
            doc_id = doc["_id"]
            del self._data[collection][doc_id]
            deleted_count += 1

        return {"deleted_count": deleted_count}

    async def count(
        self,
        collection: str,
        filter: t.Optional[t.Dict[str, t.Any]] = None,
        **kwargs: t.Any,
    ) -> int:
        docs = await self.find(collection, filter or {})
        return len(docs)

    async def aggregate(
        self,
        collection: str,
        pipeline: t.List[t.Dict[str, t.Any]],
        **kwargs: t.Any,
    ) -> t.List[t.Dict[str, t.Any]]:
        docs = await self.find(collection, {})

        for stage in pipeline:
            if "$match" in stage:
                docs = [
                    doc for doc in docs if self._matches_filter(doc, stage["$match"])
                ]
            elif "$project" in stage:
                projection = stage["$project"]
                docs = [{k: doc[k] for k in projection if k in doc} for doc in docs]
            elif "$limit" in stage:
                docs = docs[: stage["$limit"]]
            elif "$skip" in stage:
                docs = docs[stage["$skip"] :]

        return docs

    @asynccontextmanager
    async def transaction(self) -> t.AsyncGenerator[None, None]:
        self._transaction = {}
        try:
            yield
        except Exception as e:
            self.logger.error(f"Transaction failed: {e}")
            raise
        finally:
            self._transaction = None

    def _matches_filter(
        self, doc: t.Dict[str, t.Any], filter: t.Dict[str, t.Any]
    ) -> bool:
        if not filter:
            return True

        for key, value in filter.items():
            if key not in doc or doc[key] != value:
                return False
        return True


class TestNosqlBaseSettings:
    def test_init(self) -> None:
        mock_config = MagicMock()
        mock_config.app.name = "test_app"

        settings = NosqlBaseSettings()
        settings.__init__(config=mock_config)

        assert settings.host == SecretStr("127.0.0.1")
        assert settings.collection_prefix == ""
        assert settings.database == "test_app"

        settings = NosqlBaseSettings(
            host=SecretStr("custom_host"),
            port=1234,
            database="custom_db",
            collection_prefix="prefix_",
        )
        settings.__init__(config=mock_config)

        assert settings.host == SecretStr("custom_host")
        assert settings.port == 1234
        assert settings.database == "custom_db"
        assert settings.collection_prefix == "prefix_"


class TestNosqlCollection:
    @pytest.fixture
    def nosql_base(self) -> MockNosqlBase:
        return MockNosqlBase()

    @pytest.fixture
    def collection(self, nosql_base: MockNosqlBase) -> NosqlCollection:
        return NosqlCollection(nosql_base, "test_collection")

    @pytest.mark.asyncio
    async def test_find(
        self, nosql_base: MockNosqlBase, collection: NosqlCollection
    ) -> None:
        with patch.object(nosql_base, "find", AsyncMock()) as mock_find:
            mock_find.return_value = [{"_id": "1", "name": "Test"}]

            result = await collection.find({"name": "Test"})

            mock_find.assert_called_once_with("test_collection", {"name": "Test"})
            assert result == [{"_id": "1", "name": "Test"}]

    @pytest.mark.asyncio
    async def test_find_one(
        self, nosql_base: MockNosqlBase, collection: NosqlCollection
    ) -> None:
        with patch.object(nosql_base, "find_one", AsyncMock()) as mock_find_one:
            mock_find_one.return_value = {"_id": "1", "name": "Test"}

            result = await collection.find_one({"name": "Test"})

            mock_find_one.assert_called_once_with("test_collection", {"name": "Test"})
            assert result == {"_id": "1", "name": "Test"}

    @pytest.mark.asyncio
    async def test_insert_one(
        self, nosql_base: MockNosqlBase, collection: NosqlCollection
    ) -> None:
        with patch.object(nosql_base, "insert_one", AsyncMock()) as mock_insert_one:
            mock_insert_one.return_value = "1"

            result = await collection.insert_one({"name": "Test"})

            mock_insert_one.assert_called_once_with("test_collection", {"name": "Test"})
            assert result == "1"

    @pytest.mark.asyncio
    async def test_insert_many(
        self, nosql_base: MockNosqlBase, collection: NosqlCollection
    ) -> None:
        with patch.object(nosql_base, "insert_many", AsyncMock()) as mock_insert_many:
            mock_insert_many.return_value = ["1", "2"]

            result = await collection.insert_many(
                [{"name": "Test1"}, {"name": "Test2"}]
            )

            mock_insert_many.assert_called_once_with(
                "test_collection", [{"name": "Test1"}, {"name": "Test2"}]
            )
            assert result == ["1", "2"]

    @pytest.mark.asyncio
    async def test_update_one(
        self, nosql_base: MockNosqlBase, collection: NosqlCollection
    ) -> None:
        with patch.object(nosql_base, "update_one", AsyncMock()) as mock_update_one:
            mock_update_one.return_value = {"modified_count": 1}

            result = await collection.update_one(
                {"_id": "1"}, {"$set": {"name": "Updated"}}
            )

            mock_update_one.assert_called_once_with(
                "test_collection", {"_id": "1"}, {"$set": {"name": "Updated"}}
            )
            assert result == {"modified_count": 1}

    @pytest.mark.asyncio
    async def test_update_many(
        self, nosql_base: MockNosqlBase, collection: NosqlCollection
    ) -> None:
        with patch.object(nosql_base, "update_many", AsyncMock()) as mock_update_many:
            mock_update_many.return_value = {"modified_count": 2}

            result = await collection.update_many(
                {"active": True}, {"$set": {"status": "updated"}}
            )

            mock_update_many.assert_called_once_with(
                "test_collection", {"active": True}, {"$set": {"status": "updated"}}
            )
            assert result == {"modified_count": 2}

    @pytest.mark.asyncio
    async def test_delete_one(
        self, nosql_base: MockNosqlBase, collection: NosqlCollection
    ) -> None:
        with patch.object(nosql_base, "delete_one", AsyncMock()) as mock_delete_one:
            mock_delete_one.return_value = {"deleted_count": 1}

            result = await collection.delete_one({"_id": "1"})

            mock_delete_one.assert_called_once_with("test_collection", {"_id": "1"})
            assert result == {"deleted_count": 1}

    @pytest.mark.asyncio
    async def test_delete_many(
        self, nosql_base: MockNosqlBase, collection: NosqlCollection
    ) -> None:
        with patch.object(nosql_base, "delete_many", AsyncMock()) as mock_delete_many:
            mock_delete_many.return_value = {"deleted_count": 2}

            result = await collection.delete_many({"active": False})

            mock_delete_many.assert_called_once_with(
                "test_collection", {"active": False}
            )
            assert result == {"deleted_count": 2}

    @pytest.mark.asyncio
    async def test_count(
        self, nosql_base: MockNosqlBase, collection: NosqlCollection
    ) -> None:
        with patch.object(nosql_base, "count", AsyncMock()) as mock_count:
            mock_count.return_value = 5

            result = await collection.count({"active": True})

            mock_count.assert_called_once_with("test_collection", {"active": True})
            assert result == 5

    @pytest.mark.asyncio
    async def test_aggregate(
        self, nosql_base: MockNosqlBase, collection: NosqlCollection
    ) -> None:
        with patch.object(nosql_base, "aggregate", AsyncMock()) as mock_aggregate:
            mock_aggregate.return_value = [{"_id": "1", "count": 5}]

            pipeline = [{"$match": {"active": True}}, {"$project": {"count": 1}}]
            result = await collection.aggregate(pipeline)

            mock_aggregate.assert_called_once_with("test_collection", pipeline)
            assert result == [{"_id": "1", "count": 5}]


class TestMongoDBSettings:
    def test_init(self) -> None:
        mock_config = MagicMock()
        mock_config.app.name = "test_app"

        settings = MongoDBSettings()
        settings.__init__(config=mock_config)

        assert settings.port == 27017
        assert settings.database == "test_app"

        assert settings.connection_string == "mongodb://127.0.0.1:27017/test_app"

        settings = MongoDBSettings(
            host="custom_host",
            port=27018,
            user="testuser",
            password="testpass",  # nosec B106  # nosec B106
            database="custom_db",
        )
        settings.__init__(config=mock_config)

        assert settings.port == 27018
        assert settings.database == "custom_db"
        assert (
            settings.connection_string
            == "mongodb://testuser:testpass@custom_host:27018/custom_db"
        )


class TestMongoDB:
    @pytest.fixture
    def nosql(self) -> MongoDBNosql:
        instance = MongoDBNosql()
        instance.config = MagicMock()
        instance.logger = MagicMock()
        instance.config.nosql = MagicMock()
        instance.config.nosql.connection_string = "mongodb://localhost:27017/test_db"
        instance.config.nosql.connection_options = {}
        instance.config.nosql.database = "test_db"
        return instance

    def test_client_property(self, nosql: MongoDBNosql) -> None:
        with patch("acb.adapters.nosql.mongodb.AsyncIOMotorClient") as mock_client:
            mock_client.return_value = MagicMock(spec=AsyncIOMotorClient)

            client = nosql.client

            mock_client.assert_called_once_with(
                "mongodb://localhost:27017/test_db", **{}
            )
            assert client == mock_client.return_value

            client2 = nosql.client
            assert client2 is client
            mock_client.assert_called_once()

    def test_db_property(self, nosql: MongoDBNosql) -> None:
        mock_db = MagicMock(spec=AsyncIOMotorDatabase)
        mock_client = MagicMock(spec=AsyncIOMotorClient)
        mock_client.__getitem__.return_value = mock_db

        with patch.object(nosql, "client", mock_client):
            db = nosql.db

            mock_client.__getitem__.assert_called_once_with("test_db")
            assert db == mock_db

            db2 = nosql.db
            assert db2 is db
            mock_client.__getitem__.assert_called_once()

    @pytest.mark.asyncio
    async def test_init(self, nosql: MongoDBNosql) -> None:
        with (
            patch(
                "acb.adapters.nosql.mongodb.init_beanie", AsyncMock()
            ) as mock_init_beanie,
            patch.object(nosql, "db", MagicMock(spec=AsyncIOMotorDatabase)),
        ):
            await nosql.init()

            mock_init_beanie.assert_called_once_with(
                database=nosql.db, document_models=[]
            )
            nosql.logger.info.assert_called_with(
                "MongoDB connection initialized successfully"
            )

    @pytest.mark.asyncio
    async def test_init_error(self, nosql: MongoDBNosql) -> None:
        with (
            patch(
                "acb.adapters.nosql.mongodb.init_beanie", AsyncMock()
            ) as mock_init_beanie,
            patch.object(nosql, "db", MagicMock(spec=AsyncIOMotorDatabase)),
        ):
            mock_init_beanie.side_effect = Exception("Connection error")

            with pytest.raises(Exception) as excinfo:
                await nosql.init()

            assert "Connection error" in str(excinfo.value)
            nosql.logger.error.assert_called_once()

    @pytest.mark.asyncio
    async def test_find(self, nosql: MongoDBNosql) -> None:
        mock_cursor = AsyncMock()
        mock_cursor.to_list.return_value = [{"_id": "1", "name": "Test"}]

        mock_collection = MagicMock()
        mock_collection.find.return_value = mock_cursor

        mock_db = MagicMock(spec=AsyncIOMotorDatabase)
        mock_db.__getitem__.return_value = mock_collection

        with patch.object(nosql, "db", mock_db):
            result = await nosql.find("users", {"name": "Test"}, limit=10)

            mock_db.__getitem__.assert_called_once_with("users")
            mock_collection.find.assert_called_once_with({"name": "Test"}, limit=10)
            mock_cursor.to_list.assert_called_once_with(length=None)
            assert result == [{"_id": "1", "name": "Test"}]

    @pytest.mark.asyncio
    async def test_find_one(self, nosql: MongoDBNosql) -> None:
        mock_collection = MagicMock()
        mock_collection.find_one.return_value = {"_id": "1", "name": "Test"}

        mock_db = MagicMock(spec=AsyncIOMotorDatabase)
        mock_db.__getitem__.return_value = mock_collection

        with patch.object(nosql, "db", mock_db):
            result = await nosql.find_one("users", {"name": "Test"})

            mock_db.__getitem__.assert_called_once_with("users")
            mock_collection.find_one.assert_called_once_with({"name": "Test"})
            assert result == {"_id": "1", "name": "Test"}

    @pytest.mark.asyncio
    async def test_insert_one(self, nosql: MongoDBNosql) -> None:
        mock_result = MagicMock()
        mock_result.inserted_id = "1"

        mock_collection = MagicMock()
        mock_collection.insert_one.return_value = mock_result

        mock_db = MagicMock(spec=AsyncIOMotorDatabase)
        mock_db.__getitem__.return_value = mock_collection

        with patch.object(nosql, "db", mock_db):
            result = await nosql.insert_one("users", {"name": "Test"})

            mock_db.__getitem__.assert_called_once_with("users")
            mock_collection.insert_one.assert_called_once_with({"name": "Test"})
            assert result == "1"

    @pytest.mark.asyncio
    async def test_transaction(self, nosql: MongoDBNosql) -> None:
        mock_session = AsyncMock()
        mock_session.start_transaction = AsyncMock()
        mock_session.end_session = AsyncMock()

        mock_client = MagicMock(spec=AsyncIOMotorClient)
        mock_client.start_session.return_value = mock_session

        with patch.object(nosql, "client", mock_client):
            assert nosql._transaction is None

            async with nosql.transaction():
                assert nosql._transaction == mock_session

            assert nosql._transaction is None
            mock_client.start_session.assert_called_once()
            mock_session.end_session.assert_called_once()

            mock_client.start_session.reset_mock()
            mock_session.end_session.reset_mock()

            with pytest.raises(ValueError):
                async with nosql.transaction():
                    raise ValueError("Test error")


class TestRedisSettings:
    def test_init(self) -> None:
        mock_config = MagicMock()
        mock_config.app.name = "test_app"

        settings = RedisSettings()
        settings.__init__(config=mock_config)

        assert settings.port == 6379
        assert settings.db == 0
        assert settings.decode_responses
        assert settings.encoding == "utf-8"

        assert settings.connection_string == "redis://@127.0.0.1:6379/0"

        settings = RedisSettings(
            host="custom_host",
            port=6380,
            password="testpass",  # nosec B106 db=1
        )
        settings.__init__(config=mock_config)

        assert settings.port == 6380
        assert settings.db == 1
        assert settings.connection_string == "redis://:testpass@custom_host:6380/1"

    def test_cache_db_validator(self) -> None:
        settings = RedisSettings(cache_db=0)
        assert settings.cache_db == 0

        settings = RedisSettings(cache_db=3)
        assert settings.cache_db == 0

        with pytest.raises(ValueError):
            RedisSettings(cache_db=1)

        with pytest.raises(ValueError):
            RedisSettings(cache_db=2)


class TestRedis:
    @pytest.fixture
    def nosql(self) -> RedisNosql:
        instance = RedisNosql()
        instance.config = MagicMock()
        instance.logger = MagicMock()
        instance.config.nosql = MagicMock()
        instance.config.nosql.connection_string = "redis://localhost:6379/0"
        instance.config.nosql.decode_responses = True
        instance.config.nosql.encoding = "utf-8"
        instance.config.nosql.collection_prefix = "test:"
        return instance

    def test_client_property(self, nosql: RedisNosql) -> None:
        with patch("acb.adapters.nosql.redis.redis.from_url") as mock_from_url:
            mock_from_url.return_value = MagicMock(spec=Redis)

            client = nosql.client

            mock_from_url.assert_called_once_with(
                "redis://localhost:6379/0", decode_responses=True, encoding="utf-8"
            )
            assert client == mock_from_url.return_value

            client2 = nosql.client
            assert client2 is client
            mock_from_url.assert_called_once()

    def test_om_client_property(self, nosql: RedisNosql) -> None:
        with patch("acb.adapters.nosql.redis.get_redis_connection") as mock_get_redis:
            mock_get_redis.return_value = MagicMock()

            client = nosql.om_client

            mock_get_redis.assert_called_once_with(
                url="redis://localhost:6379/0", decode_responses=True
            )
            assert client == mock_get_redis.return_value

            client2 = nosql.om_client
            assert client2 is client
            mock_get_redis.assert_called_once()

    @pytest.mark.asyncio
    async def test_init(self, nosql: RedisNosql) -> None:
        with (
            patch.object(nosql, "client", MagicMock(spec=Redis)) as mock_client,
            patch("acb.adapters.nosql.redis.Migrator") as mock_migrator_class,
        ):
            mock_client.ping = AsyncMock()
            mock_migrator = MagicMock()
            mock_migrator_class.return_value = mock_migrator

            await nosql.init()

            mock_client.ping.assert_called_once()
            mock_migrator_class.assert_called_once()
            mock_migrator.run.assert_called_once()
            nosql.logger.info.assert_called_with(
                "Redis connection initialized successfully"
            )

    @pytest.mark.asyncio
    async def test_init_error(self, nosql: RedisNosql) -> None:
        with patch.object(nosql, "client", MagicMock(spec=Redis)) as mock_client:
            mock_client.ping = AsyncMock(side_effect=Exception("Connection error"))

            with pytest.raises(Exception) as excinfo:
                await nosql.init()

            assert "Connection error" in str(excinfo.value)
            nosql.logger.error.assert_called_once()

    def test_get_key(self, nosql: RedisNosql) -> None:
        key = nosql._get_key("users", "123")
        assert key == "test:users:123"

        key = nosql._get_key("users")
        assert key == "test:users"

    def test_matches_filter(self, nosql: RedisNosql) -> None:
        assert nosql._matches_filter({"name": "Test"}, {})

        assert nosql._matches_filter({"name": "Test", "age": 30}, {"name": "Test"})

        assert not nosql._matches_filter({"name": "Test"}, {"name": "Other"})

        assert not nosql._matches_filter({"name": "Test"}, {"age": 30})

    @pytest.mark.asyncio
    async def test_transaction(self, nosql: RedisNosql) -> None:
        with patch.object(nosql, "client", MagicMock(spec=Redis)) as mock_client:
            mock_pipeline = AsyncMock()
            mock_client.pipeline.return_value = mock_pipeline

            assert nosql._transaction is None

            async with nosql.transaction():
                assert nosql._transaction == mock_pipeline

            assert nosql._transaction is None
            mock_client.pipeline.assert_called_once()
            mock_pipeline.execute.assert_called_once()

            mock_client.pipeline.reset_mock()
            mock_pipeline.execute.reset_mock()

            with pytest.raises(ValueError):
                async with nosql.transaction():
                    raise ValueError("Test error")


class TestFirestoreSettings:
    def test_init(self) -> None:
        mock_config = MagicMock()
        mock_config.app.name = "test_app"
        mock_config.app.project = "test-project"

        settings = FirestoreSettings()
        settings.__init__(config=mock_config)

        assert settings.project_id == "test-project"

        settings = FirestoreSettings(
            project_id="custom-project",
            credentials_path="/path/to/credentials.json",
            emulator_host="localhost:8080",
        )
        settings.__init__(config=mock_config)

        assert settings.project_id == "custom-project"
        assert settings.credentials_path == "/path/to/credentials.json"
        assert settings.emulator_host == "localhost:8080"


class TestFirestore:
    @pytest.fixture
    def nosql(self) -> FirestoreNosql:
        instance = FirestoreNosql()
        instance.config = MagicMock()
        instance.logger = MagicMock()
        instance.config.nosql = MagicMock()
        instance.config.nosql.project_id = "test-project"
        instance.config.nosql.credentials_path = None
        instance.config.nosql.collection_prefix = "test_"
        return instance

    def test_client_property(self, nosql: FirestoreNosql) -> None:
        with patch(
            "acb.adapters.nosql.firestore.firestore.Client"
        ) as mock_client_class:
            mock_client = MagicMock(spec=firestore.Client)
            mock_client_class.return_value = mock_client

            client = nosql.client

            mock_client_class.assert_called_once_with(project="test-project")
            assert client == mock_client

            client2 = nosql.client
            assert client2 is client
            mock_client_class.assert_called_once()

    def test_db_property(self, nosql: FirestoreNosql) -> None:
        with patch.object(
            nosql, "client", MagicMock(spec=firestore.Client)
        ) as mock_client:
            db = nosql.db

            assert db == mock_client

    @pytest.mark.asyncio
    async def test_init(self, nosql: FirestoreNosql) -> None:
        with patch.object(
            nosql, "client", MagicMock(spec=firestore.Client)
        ) as mock_client:
            mock_client.collection.return_value = MagicMock()

            await nosql.init()

            mock_client.collection.assert_called_once_with("test")
            nosql.logger.info.assert_called_with(
                "Firestore connection initialized successfully"
            )

    @pytest.mark.asyncio
    async def test_init_error(self, nosql: FirestoreNosql) -> None:
        with patch.object(
            nosql, "client", MagicMock(spec=firestore.Client)
        ) as mock_client:
            mock_client.collection.side_effect = Exception("Connection error")

            with pytest.raises(Exception) as excinfo:
                await nosql.init()

            assert "Connection error" in str(excinfo.value)
            nosql.logger.error.assert_called_once()

    def test_get_collection_ref(self, nosql: FirestoreNosql) -> None:
        with patch.object(
            nosql, "client", MagicMock(spec=firestore.Client)
        ) as mock_client:
            mock_collection = MagicMock(spec=firestore.CollectionReference)
            mock_client.collection.return_value = mock_collection

            collection_ref = nosql._get_collection_ref("users")

            mock_client.collection.assert_called_once_with("test_users")
            assert collection_ref == mock_collection

    def test_convert_to_dict(self, nosql: FirestoreNosql) -> None:
        mock_doc = MagicMock(spec=firestore.DocumentSnapshot)
        mock_doc.exists = True
        mock_doc.id = "123"
        mock_doc.to_dict.return_value = {"name": "Test", "age": 30}



import os
import typing as t
from contextlib import asynccontextmanager
from pathlib import Path
from types import TracebackType
from typing import Optional, Type
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from pydantic import SecretStr
from acb.adapters.ftpd._base import FileInfo, FtpdBase, FtpdBaseSettings
from acb.adapters.ftpd.ftp import Ftpd as FTPAdapter
from acb.adapters.ftpd.ftp import FtpdSettings as FTPSettings
from acb.adapters.ftpd.sftp import Ftpd as SFTPAdapter
from acb.adapters.ftpd.sftp import FtpdSettings as SFTPSettings


class MockFtpdBase(FtpdBase):
    def __init__(self) -> None:
        self.config = MagicMock()
        self.logger = MagicMock()
        self._files: dict[str, bytes] = {}
        self._dirs: set[str] = {"/"}

    async def start(self) -> None:
        self.logger.info("Mock FTP/SFTP server started")

    async def stop(self) -> None:
        self.logger.info("Mock FTP/SFTP server stopped")

    async def upload(self, local_path: Path, remote_path: str) -> None:
        mock_content = b"Mock file content"
        await self.write_bytes(remote_path, mock_content)

    async def download(self, remote_path: str, local_path: Path) -> None:
        await self.read_bytes(remote_path)

    async def list_dir(self, path: str) -> t.List[FileInfo]:
        if path not in self._dirs:
            raise FileNotFoundError(f"Directory not found: {path}")

        result = []
        for file_path, content in self._files.items():
            dir_name = os.path.dirname(file_path)
            if dir_name == path:
                file_name = os.path.basename(file_path)
                info = FileInfo(name=file_name, size=len(content))
                result.append(info)

        for dir_path in self._dirs:
            if dir_path != path and os.path.dirname(dir_path) == path:
                dir_name = os.path.basename(dir_path)
                info = FileInfo(name=dir_name, is_dir=True, is_file=False)
                result.append(info)

        return result

    async def mkdir(self, path: str) -> None:
        self._dirs.add(path)
        parent = os.path.dirname(path)
        if parent and parent != "/" and parent not in self._dirs:
            await self.mkdir(parent)

    async def rmdir(self, path: str, recursive: bool = False) -> None:
        if path not in self._dirs:
            raise FileNotFoundError(f"Directory not found: {path}")

        if recursive:
            for file_path in list(self._files.keys()):
                if os.path.dirname(file_path) == path:
                    del self._files[file_path]

            for dir_path in list(self._dirs):
                if dir_path != path and dir_path.startswith(path + "/"):
                    self._dirs.remove(dir_path)

        self._dirs.remove(path)

    async def delete(self, path: str) -> None:
        if path not in self._files:
            raise FileNotFoundError(f"File not found: {path}")
        del self._files[path]

    async def rename(self, old_path: str, new_path: str) -> None:
        if old_path in self._files:
            self._files[new_path] = self._files[old_path]
            del self._files[old_path]
        elif old_path in self._dirs:
            self._dirs.add(new_path)
            self._dirs.remove(old_path)

            for file_path in list(self._files.keys()):
                if file_path.startswith(old_path + "/"):
                    new_file_path = new_path + file_path[len(old_path) :]
                    self._files[new_file_path] = self._files[file_path]
                    del self._files[file_path]

            for dir_path in list(self._dirs):
                if dir_path.startswith(old_path + "/"):
                    new_dir_path = new_path + dir_path[len(old_path) :]
                    self._dirs.add(new_dir_path)
                    self._dirs.remove(dir_path)
        else:
            raise FileNotFoundError(f"Path not found: {old_path}")

    async def exists(self, path: str) -> bool:
        return path in self._files or path in self._dirs

    async def stat(self, path: str) -> FileInfo:
        if path in self._files:
            return FileInfo(
                name=os.path.basename(path),
                size=len(self._files[path]),
            )
        elif path in self._dirs:
            return FileInfo(
                name=os.path.basename(path) or path, is_dir=True, is_file=False
            )
        else:
            raise FileNotFoundError(f"Path not found: {path}")

    async def read_text(self, path: str) -> str:
        return (await self.read_bytes(path)).decode()

    async def read_bytes(self, path: str) -> bytes:
        if path not in self._files:
            raise FileNotFoundError(f"File not found: {path}")
        return self._files[path]

    async def write_text(self, path: str, content: str) -> None:
        await self.write_bytes(path, content.encode())

    async def write_bytes(self, path: str, content: bytes) -> None:
        parent = os.path.dirname(path)
        if parent and parent != "/" and parent not in self._dirs:
            await self.mkdir(parent)
        self._files[path] = content

    async def __aenter__(self) -> t.Any:
        return self

    async def __aexit__(
        self,
        _exc_type: Optional[Type[BaseException]],
        _exc_val: Optional[BaseException],
        _exc_tb: Optional[TracebackType],
    ) -> None:
        pass

    @asynccontextmanager
    async def connect(self) -> "t.AsyncGenerator[FtpdBase, None]":
        yield self


class TestFtpdBaseSettings:
    def test_init(self) -> None:
        settings = FtpdBaseSettings()

        assert settings.host == "127.0.0.1"
        assert settings.port == 8021
        assert settings.max_connections == 42
        assert settings.username == "ftpuser"
        assert settings.password.get_secret_value() == "ftppass"
        assert settings.anonymous is False
        assert settings.root_dir == "/tmp/ftp"  # nosec B108
        assert settings.use_tls is False
        assert settings.cert_file is None
        assert settings.key_file is None

        settings = FtpdBaseSettings(
            host="custom_host",
            port=2121,
            max_connections=100,
            username="custom_user",
            password=SecretStr("custom_pass"),
            anonymous=True,
            root_dir="/custom/path",
            use_tls=True,
            cert_file="/path/to/cert.pem",
            key_file="/path/to/key.pem",
        )

        assert settings.host == "custom_host"
        assert settings.port == 2121
        assert settings.max_connections == 100
        assert settings.username == "custom_user"
        assert settings.password.get_secret_value() == "custom_pass"
        assert settings.anonymous is True
        assert settings.root_dir == "/custom/path"
        assert settings.use_tls is True
        assert settings.cert_file == "/path/to/cert.pem"
        assert settings.key_file == "/path/to/key.pem"


class TestFileInfo:
    def test_init(self) -> None:
        info = FileInfo(name="test.txt")
        assert info.name == "test.txt"
        assert info.size == 0
        assert not info.is_dir
        assert info.is_file
        assert not info.is_symlink
        assert info.permissions == ""
        assert info.mtime == 0.0
        assert info.owner == ""
        assert info.group == ""

        info = FileInfo(
            name="test.txt",
            size=1024,
            is_dir=True,
            is_file=False,
            is_symlink=True,
            permissions="rwxr-xr-x",
            mtime=1234567890.0,
            owner="user",
            group="group",
        )
        assert info.name == "test.txt"
        assert info.size == 1024
        assert info.is_dir
        assert not info.is_file
        assert info.is_symlink
        assert info.permissions == "rwxr-xr-x"
        assert info.mtime == 1234567890.0
        assert info.owner == "user"
        assert info.group == "group"


class TestMockFtpdBase:
    @pytest.fixture
    def ftpd(self) -> MockFtpdBase:
        return MockFtpdBase()

    @pytest.mark.asyncio
    async def test_start_stop(self, ftpd: MockFtpdBase) -> None:
        await ftpd.start()
        ftpd.logger.info.assert_called_with("Mock FTP/SFTP server started")

        await ftpd.stop()
        ftpd.logger.info.assert_called_with("Mock FTP/SFTP server stopped")

    @pytest.mark.asyncio
    async def test_file_operations(self, ftpd: MockFtpdBase) -> None:
        await ftpd.write_text("/test.txt", "Hello, world!")

        assert await ftpd.exists("/test.txt")

        content = await ftpd.read_text("/test.txt")
        assert content == "Hello, world!"

        info = await ftpd.stat("/test.txt")
        assert info.name == "test.txt"
        assert info.size == 13
        assert info.is_file
        assert not info.is_dir

        await ftpd.rename("/test.txt", "/renamed.txt")
        assert not await ftpd.exists("/test.txt")
        assert await ftpd.exists("/renamed.txt")

        await ftpd.delete("/renamed.txt")
        assert not await ftpd.exists("/renamed.txt")

    @pytest.mark.asyncio
    async def test_directory_operations(self, ftpd: MockFtpdBase) -> None:
        await ftpd.mkdir("/testdir")

        assert await ftpd.exists("/testdir")

        info = await ftpd.stat("/testdir")
        assert info.name == "testdir"
        assert info.is_dir
        assert not info.is_file

        await ftpd.write_text("/testdir/test.txt", "Hello, world!")

        files = await ftpd.list_dir("/")
        assert len(files) == 1
        assert files[0].name == "testdir"
        assert files[0].is_dir

        files = await ftpd.list_dir("/testdir")
        assert len(files) == 1
        assert files[0].name == "test.txt"
        assert files[0].is_file

        await ftpd.rename("/testdir", "/renamed")
        assert not await ftpd.exists("/testdir")
        assert await ftpd.exists("/renamed")
        assert await ftpd.exists("/renamed/test.txt")

        with pytest.raises(Exception):
            await ftpd.rmdir("/renamed")

        await ftpd.rmdir("/renamed", recursive=True)
        assert not await ftpd.exists("/renamed")

    @pytest.mark.asyncio
    async def test_upload_download(self, ftpd: MockFtpdBase) -> None:
        mock_local_path = MagicMock(spec=Path)

        await ftpd.upload(mock_local_path, "/uploaded.txt")

        assert await ftpd.exists("/uploaded.txt")

        content = await ftpd.read_text("/uploaded.txt")
        assert content is not None

        mock_download_path = MagicMock(spec=Path)

        await ftpd.download("/uploaded.txt", mock_download_path)


class TestFTPAdapter:
    @pytest.fixture
    def ftpd(self) -> FTPAdapter:
        adapter = FTPAdapter()
        adapter.config = MagicMock()
        adapter.config.ftpd = FTPSettings(
            host="127.0.0.1",
            port=8021,
            username="ftpuser",
            password=SecretStr("ftppass"),
            root_dir="/mock/ftp",
        )
        adapter.logger = MagicMock()
        return adapter

    def test_server_property(self, ftpd: FTPAdapter) -> None:
        with patch("os.makedirs") as mock_makedirs:
            server = ftpd.server

            mock_makedirs.assert_called_once_with("/mock/ftp", exist_ok=True)
            assert server is not None

            server2 = ftpd.server
            assert server2 is server

    @pytest.mark.asyncio
    async def test_start(self, ftpd: FTPAdapter) -> None:
        with patch.object(ftpd, "server") as mock_server:
            mock_server.start = AsyncMock()

            await ftpd.start()

            mock_server.start.assert_called_once()
            ftpd.logger.info.assert_called_once()

    @pytest.mark.asyncio
    async def test_stop(self, ftpd: FTPAdapter) -> None:
        with patch.object(ftpd, "server") as mock_server:
            mock_server.close = AsyncMock()

            await ftpd.stop()

            mock_server.close.assert_called_once()
            ftpd.logger.info.assert_called_once()


class TestSFTPAdapter:
    @pytest.fixture
    def ftpd(self) -> SFTPAdapter:
        adapter = SFTPAdapter()
        adapter.config = MagicMock()
        adapter.config.ftpd = SFTPSettings(
            host="127.0.0.1",
            port=8022,
            username="ftpuser",
            password=SecretStr("ftppass"),
            root_dir="/mock/sftp",
            server_host_keys=["ssh_host_key"],
            authorized_client_keys="authorized_keys",
        )
        adapter.logger = MagicMock()
        return adapter

    def test_server_factory_property(self, ftpd: SFTPAdapter) -> None:
        factory = ftpd.server_factory
        assert factory is not None

        factory2 = ftpd.server_factory
        assert factory2 is factory

    @pytest.mark.asyncio
    async def test_start(self, ftpd: SFTPAdapter) -> None:
        with (
            patch("os.makedirs") as mock_makedirs,
            patch(
                "asyncssh.create_server", new_callable=AsyncMock
            ) as mock_create_server,
        ):
            mock_server = MagicMock()
            mock_create_server.return_value = mock_server

            await ftpd.start()

            mock_makedirs.assert_called_once_with("/mock/sftp", exist_ok=True)
            mock_create_server.assert_called_once()
            ftpd.logger.info.assert_called_once()
            assert ftpd._server is mock_server

    @pytest.mark.asyncio
    async def test_stop(self, ftpd: SFTPAdapter) -> None:
        mock_server = MagicMock()
        mock_server.close = MagicMock()
        mock_server.wait_closed = AsyncMock()
        ftpd._server = mock_server

        await ftpd.stop()

        mock_server.close.assert_called_once()
        mock_server.wait_closed.assert_called_once()
        ftpd.logger.info.assert_called_once()
        assert ftpd._server is None

    def test_create_server_connection(self, ftpd: SFTPAdapter) -> None:
        connection = ftpd._create_server_connection()
        assert connection is not None

    def test_process_factory(self, ftpd: SFTPAdapter) -> None:
        mock_process = MagicMock()
        mock_process.exit = MagicMock()

        ftpd._process_factory(mock_process)

        mock_process.exit.assert_called_once_with(0)



from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from httpx import Response as HttpxResponse
from pydantic import SecretStr
from acb.adapters.dns._base import DnsRecord
from acb.adapters.smtp._base import SmtpBase, SmtpBaseSettings
from acb.adapters.smtp.gmail import Smtp as GmailSmtp
from acb.adapters.smtp.gmail import SmtpSettings as GmailSmtpSettings
from acb.adapters.smtp.mailgun import (
    Smtp as MailgunSmtp,
)
from acb.adapters.smtp.mailgun import (
    SmtpSettings as MailgunSmtpSettings,
)


class TestSmtpBaseSettings:
    def test_init(self) -> None:
        mock_config = MagicMock()
        mock_config.app = MagicMock()
        mock_config.app.domain = "example.com"
        mock_config.app.title = "Example App"

        settings = SmtpBaseSettings()

        settings.__init__(config=mock_config)

        assert settings.domain == "mail.example.com"
        assert settings.default_from == "info@example.com"
        assert settings.default_from_name == "Example App"
        assert settings.port == 587
        assert settings.tls
        assert not settings.ssl
        assert isinstance(settings.password, SecretStr)
        assert "admin" in settings.forwards
        assert settings.forwards["admin"] == "pat@example.com"

        settings = SmtpBaseSettings()

        settings.domain = "custom.example.com"
        settings.default_from = "custom@example.com"
        settings.default_from_name = "Custom App"
        settings.port = 465
        settings.tls = False
        settings.ssl = True
        settings.forwards = {"support": "support@example.com"}

        assert settings.domain == "custom.example.com"
        assert settings.default_from == "custom@example.com"
        assert settings.default_from_name == "Custom App"
        assert settings.port == 465
        assert not settings.tls
        assert settings.ssl
        assert settings.forwards["support"] == "support@example.com"


class MockSmtpBase(SmtpBase):
    def __init__(self) -> None:
        self.config = MagicMock()
        self.logger = MagicMock()
        self.initialized = False

    async def init(self) -> None:
        self.initialized = True


class TestSmtpBase:
    @pytest.fixture
    def smtp_base(self) -> MockSmtpBase:
        return MockSmtpBase()

    @pytest.mark.asyncio
    async def test_init(self, smtp_base: MockSmtpBase) -> None:
        await smtp_base.init()
        assert smtp_base.initialized


class TestMailgunSmtpSettings:
    def test_init(self) -> None:
        settings = MailgunSmtpSettings()

        settings.domain = "mail.example.com"
        settings.default_from = "info@example.com"
        settings.default_from_name = "Example App"
        settings.api_url = "https://api.mailgun.net/v3/domains"
        settings.mx_servers = ["smtp.mailgun.com"]
        settings.api_key = SecretStr("test-api-key")
        settings.password = SecretStr("test-password")

        assert settings.domain == "mail.example.com"
        assert settings.default_from == "info@example.com"
        assert settings.default_from_name == "Example App"
        assert settings.api_url == "https://api.mailgun.net/v3/domains"
        assert settings.mx_servers == ["smtp.mailgun.com"]
        assert settings.api_key.get_secret_value() == "test-api-key"
        assert settings.password.get_secret_value() == "test-password"


class TestMailgunSmtp:
    @pytest.fixture
    def smtp(self) -> MailgunSmtp:
        instance = MailgunSmtp()
        instance.config = MagicMock()
        instance.config.smtp = MagicMock()
        instance.config.smtp.api_url = "https://api.mailgun.net/v3/domains"
        instance.config.smtp.api_key = MagicMock()
        instance.config.smtp.api_key.get_secret_value.return_value = "test-api-key"
        instance.config.smtp.domain = "example.com"
        instance.config.app = MagicMock()
        instance.config.app.domain = "example.com"
        instance.config.smtp.forwards = {"admin": "admin@example.com"}
        instance.logger = MagicMock()
        instance.requests = MagicMock()
        return instance

    @pytest.mark.asyncio
    async def test_get_response_get(self, smtp: MailgunSmtp) -> None:
        mock_response = MagicMock(spec=HttpxResponse)
        mock_response.json.return_value = {"message": "success"}
        smtp.requests.get = AsyncMock(return_value=mock_response)

        with patch(
            "acb.adapters.smtp.mailgun.load.json",
            AsyncMock(return_value={"message": "success"}),
        ):
            result = await smtp.get_response(
                "get", domain="example.com", params={"limit": 10}
            )

            smtp.requests.get.assert_called_once()
            assert result == {"message": "success"}

    @pytest.mark.asyncio
    async def test_get_response_put(self, smtp: MailgunSmtp) -> None:
        mock_response = MagicMock(spec=HttpxResponse)
        mock_response.json.return_value = {"message": "success"}
        smtp.requests.put = AsyncMock(return_value=mock_response)

        with (
            patch(
                "acb.adapters.smtp.mailgun.load.json",
                AsyncMock(return_value={"message": "success"}),
            ),
            patch("acb.adapters.smtp.mailgun.sys._getframe") as mock_frame,
        ):
            mock_frame.return_value.f_back.f_code.co_name = "test_domain"

            result = await smtp.get_response(
                "put", domain="example.com", data={"key": "value"}
            )

            smtp.requests.put.assert_called_once()
            assert result == {"message": "success"}

    @pytest.mark.asyncio
    async def test_list_domains(self, smtp: MailgunSmtp) -> None:
        with patch.object(smtp, "get_response", AsyncMock()) as mock_get_response:
            mock_get_response.return_value = {
                "items": [{"name": "example.com"}, {"name": "test.com"}]
            }

            domains = await smtp.list_domains()

            mock_get_response.assert_called_once_with(
                "get", params={"skip": 0, "limit": 1000}
            )
            assert domains == ["example.com", "test.com"]

    @pytest.mark.asyncio
    async def test_get_domain(self, smtp: MailgunSmtp) -> None:
        with patch.object(smtp, "get_response", AsyncMock()) as mock_get_response:
            mock_get_response.return_value = {
                "domain": "example.com",
                "state": "active",
            }

            domain_info = await smtp.get_domain("example.com")

            mock_get_response.assert_called_once_with("get", domain="example.com")
            assert domain_info == {"domain": "example.com", "state": "active"}

    @pytest.mark.asyncio
    async def test_create_domain(self, smtp: MailgunSmtp) -> None:
        with patch.object(smtp, "get_response", AsyncMock()) as mock_get_response:
            mock_get_response.side_effect = [
                {"message": "Domain created"},
                {"message": "Connection updated"},
            ]

            result = await smtp.create_domain("example.com")

            assert mock_get_response.call_count == 2
            assert result == {"message": "Connection updated"}

    @pytest.mark.asyncio
    async def test_delete_domain(self, smtp: MailgunSmtp) -> None:
        with patch.object(smtp, "get_response", AsyncMock()) as mock_get_response:
            mock_get_response.return_value = {"message": "Domain deleted"}

            result = await smtp.delete_domain("example.com")

            mock_get_response.assert_called_once_with("delete", domain="example.com")
            assert result == {"message": "Domain deleted"}

    @pytest.mark.asyncio
    async def test_get_dns_records(self, smtp: MailgunSmtp) -> None:
        with (
            patch.object(smtp, "get_domain", AsyncMock()) as mock_get_domain,
            patch("acb.adapters.smtp.mailgun.debug") as mock_debug,
        ):
            mock_get_domain.return_value = {
                "receiving_dns_records": [
                    {"priority": "10", "value": "mx.example.com"}
                ],
                "sending_dns_records": [
                    {"name": "example.com", "record_type": "TXT", "value": "v=spf1"}
                ],
            }

            records = await smtp.get_dns_records("example.com")

            mock_get_domain.assert_called_once_with("example.com")
            assert len(records) == 2
            assert records[0].type == "MX"
            assert records[1].type == "TXT"
            mock_debug.assert_called_once()

    @pytest.mark.asyncio
    async def test_create_dns_records(self, smtp: MailgunSmtp) -> None:
        mock_dns = AsyncMock()

        with (
            patch.object(smtp, "create_domain", AsyncMock()) as mock_create_domain,
            patch.object(
                smtp, "create_domain_credentials", AsyncMock()
            ) as mock_create_creds,
            patch.object(smtp, "get_dns_records", AsyncMock()) as mock_get_records,
            patch.object(smtp, "delete_domain", AsyncMock()) as mock_delete_domain,
        ):
            mock_records = [MagicMock(spec=DnsRecord), MagicMock(spec=DnsRecord)]
            mock_get_records.return_value = mock_records
            smtp.config.mail.mailgun.gmail.enabled = False

            await smtp.create_dns_records(dns=mock_dns)

            mock_create_domain.assert_called_once_with(smtp.config.smtp.domain)
            mock_create_creds.assert_called_once_with(smtp.config.smtp.domain)
            mock_get_records.assert_called_once_with(smtp.config.smtp.domain)
            mock_delete_domain.assert_called_once_with(smtp.config.smtp.domain)
            mock_dns.create_records.assert_called_once_with(mock_records)

    @pytest.mark.asyncio
    async def test_create_route(self, smtp: MailgunSmtp) -> None:
        with (
            patch.object(smtp, "list_routes", AsyncMock()) as mock_list_routes,
            patch.object(smtp, "get_response", AsyncMock()) as mock_get_response,
        ):
            mock_list_routes.return_value = []
            mock_get_response.return_value = {"message": "Route created"}

            result = await smtp.create_route("admin", "admin@example.com")

            mock_list_routes.assert_called_once()
            mock_get_response.assert_called_once()
            assert result == {"message": "Route created"}
            smtp.logger.info.assert_called_once()

    @pytest.mark.asyncio
    async def test_init(self, smtp: MailgunSmtp) -> None:
        with (
            patch.object(smtp, "create_dns_records", AsyncMock()) as mock_create_dns,
            patch.object(smtp, "delete_routes", AsyncMock()) as mock_delete_routes,
            patch.object(smtp, "create_routes", AsyncMock()) as mock_create_routes,
        ):
            await smtp.init()

            mock_create_dns.assert_called_once()
            mock_delete_routes.assert_called_once()
            mock_create_routes.assert_called_once()


class TestGmailSmtpSettings:
    def test_init(self) -> None:
        settings = GmailSmtpSettings()

        settings.domain = "mail.example.com"
        settings.default_from = "info@example.com"
        settings.default_from_name = "Example App"
        settings.token_uri = "https://oauth2.googleapis.com/token"
        settings.mx_servers = [
            "1 aspmx.l.google.com.",
            "2 alt1.aspmx.l.google.com.",
            "3 alt2.aspmx.l.google.com.",
            "4 alt3.aspmx.l.google.com.",
            "5 alt4.aspmx.l.google.com.",
        ]

        assert settings.domain == "mail.example.com"
        assert settings.default_from == "info@example.com"
        assert settings.default_from_name == "Example App"
        assert settings.token_uri == "https://oauth2.googleapis.com/token"
        assert "https://www.googleapis.com/auth/gmail.send" in settings.scopes
        assert settings.mx_servers is not None and len(settings.mx_servers) == 5
        assert (
            settings.mx_servers is not None
            and "1 aspmx.l.google.com." in settings.mx_servers
        )

        settings = GmailSmtpSettings()
        settings.client_id = "test-client-id"
        settings.client_secret = SecretStr("test-client-secret")
        settings.refresh_token = SecretStr("test-refresh-token")

        assert settings.client_id == "test-client-id"
        assert settings.client_secret.get_secret_value() == "test-client-secret"
        assert settings.refresh_token.get_secret_value() == "test-refresh-token"


class TestGmailSmtp:
    @pytest.fixture
    def smtp(self) -> GmailSmtp:
        instance = GmailSmtp()
        instance.config = MagicMock()
        instance.config.smtp = MagicMock()
        instance.config.smtp.domain = "example.com"
        instance.config.smtp.default_from = "info@example.com"
        instance.config.smtp.default_from_name = "Example App"
        instance.config.smtp.client_id = "test-client-id"
        instance.config.smtp.client_secret = MagicMock()
        instance.config.smtp.client_secret.get_secret_value.return_value = (
            "test-client-secret"
        )
        instance.config.smtp.refresh_token = MagicMock()
        instance.config.smtp.refresh_token.get_secret_value.return_value = (
            "test-refresh-token"
        )
        instance.config.smtp.token_uri = "https://oauth2.googleapis.com/token"
        instance.config.smtp.scopes = ["https://www.googleapis.com/auth/gmail.send"]
        instance.config.smtp.mx_servers = [
            "1 aspmx.l.google.com.",
            "5 alt1.aspmx.l.google.com.",
        ]
        instance.config.smtp.forwards = {"admin": "admin@example.com"}
        instance.logger = MagicMock()
        instance.requests = MagicMock()
        return instance

    def test_get_gmail_service(self, smtp: GmailSmtp) -> None:
        with (
            patch("acb.adapters.smtp.gmail.Credentials") as mock_credentials,
            patch("acb.adapters.smtp.gmail.build") as mock_build,
        ):
            mock_service = MagicMock()
            mock_build.return_value = mock_service

            service = smtp._get_gmail_service()

            mock_credentials.assert_called_once_with(
                None,
                refresh_token="test-refresh-token",
                token_uri="https://oauth2.googleapis.com/token",
                client_id="test-client-id",
                client_secret="test-client-secret",
                scopes=["https://www.googleapis.com/auth/gmail.send"],
            )
            mock_build.assert_called_once_with(
                "gmail", "v1", credentials=mock_credentials.return_value
            )
            assert service == mock_service

    @pytest.mark.asyncio
    async def test_get_response(self, smtp: GmailSmtp) -> None:
        result = await smtp.get_response("get", domain="example.com")

        smtp.logger.debug.assert_called_once()
        assert result["message"] == "success"
        assert result["status"] == "ok"

    @pytest.mark.asyncio
    async def test_list_domains(self, smtp: GmailSmtp) -> None:
        domains = await smtp.list_domains()

        assert domains == ["example.com"]

    @pytest.mark.asyncio
    async def test_get_domain(self, smtp: GmailSmtp) -> None:
        domain_info = await smtp.get_domain("example.com")
        assert domain_info is not None and domain_info["domain"] == "example.com"
        assert (
            domain_info is not None and len(domain_info["receiving_dns_records"]) == 2
        )
        assert (
            domain_info is not None
            and domain_info["receiving_dns_records"][0]["priority"] == "1"
        )
        assert (
            domain_info is not None
            and domain_info["receiving_dns_records"][0]["value"]
            == "aspmx.l.google.com."
        )
        assert domain_info is not None and len(domain_info["sending_dns_records"]) == 2
        assert (
            domain_info is not None
            and "v=spf1" in domain_info["sending_dns_records"][0]["value"]
        )
        assert (
            domain_info is not None
            and "v=DMARC1" in domain_info["sending_dns_records"][1]["value"]
        )

    @pytest.mark.asyncio
    async def test_create_domain(self, smtp: GmailSmtp) -> None:
        result = await smtp.create_domain("example.com")

        smtp.logger.info.assert_called_once()
        assert result["message"] == "Domain configured for Gmail"
        assert result["domain"] == "example.com"

    @pytest.mark.asyncio
    async def test_delete_domain(self, smtp: GmailSmtp) -> None:
        result = await smtp.delete_domain("example.com")

        smtp.logger.info.assert_called_once()
        assert result["message"] == "Domain deletion simulated"
        assert result["domain"] == "example.com"

    @pytest.mark.asyncio
    async def test_get_dns_records(self, smtp: GmailSmtp) -> None:
        with patch("acb.adapters.smtp.gmail.debug") as mock_debug:
            records = await smtp.get_dns_records("example.com")

            assert len(records) == 3
            assert records[0].type == "MX"
            assert records[1].type == "TXT"
            assert records[2].type == "TXT"
            mock_debug.assert_called_once()

    @pytest.mark.asyncio
    async def test_create_dns_records(self, smtp: GmailSmtp) -> None:
        mock_dns = AsyncMock()

        with patch.object(smtp, "get_dns_records", AsyncMock()) as mock_get_records:
            mock_records = [MagicMock(), MagicMock()]
            mock_get_records.return_value = mock_records

            await smtp.create_dns_records(dns=mock_dns)

            mock_get_records.assert_called_once_with("example.com")
            mock_dns.create_records.assert_called_once_with(mock_records)
            smtp.logger.info.assert_called_once()

    @pytest.mark.asyncio
    async def test_list_routes(self, smtp: GmailSmtp) -> None:
        with patch.object(smtp, "_get_gmail_service") as mock_get_service:
            mock_service = MagicMock()
            mock_get_service.return_value = mock_service

            mock_list = MagicMock()
            mock_service.users().settings().forwardingAddresses().list.return_value = (
                mock_list
            )
            mock_list.execute.return_value = {
                "forwardingAddresses": [{"forwardingEmail": "admin@example.com"}]
            }

            routes = await smtp.list_routes()

            assert len(routes) == 1
            assert routes[0]["id"] == "admin@example.com"
            assert "match_recipient" in routes[0]["expression"]
            assert "forward" in routes[0]["actions"][0]

    @pytest.mark.asyncio
    async def test_delete_route(self, smtp: GmailSmtp) -> None:
        with patch.object(smtp, "_get_gmail_service") as mock_get_service:
            mock_service = MagicMock()
            mock_get_service.return_value = mock_service

            mock_delete = MagicMock()
            mock_service.users().settings().forwardingAddresses().delete.return_value = mock_delete

            route = {"id": "admin@example.com", "description": "Admin route"}
            response = await smtp.delete_route(route)

            mock_service.users().settings().forwardingAddresses().delete.assert_called_once_with(
                userId="me", forwardingEmail="admin@example.com"
            )
            assert response.status_code == 200
            smtp.logger.info.assert_called_once()

    @pytest.mark.asyncio
    async def test_create_route(self, smtp: GmailSmtp) -> None:
        with patch.object(smtp, "_get_gmail_service") as mock_get_service:
            mock_service = MagicMock()
            mock_get_service.return_value = mock_service

            mock_create = MagicMock()
            mock_service.users().settings().forwardingAddresses().create.return_value = mock_create

            mock_update = MagicMock()
            mock_service.users().settings().updateAutoForwarding.return_value = (
                mock_update
            )

            result = await smtp.create_route("admin", "admin@example.com")

            mock_service.users().settings().forwardingAddresses().create.assert_called_once_with(
                userId="me", body={"forwardingEmail": "admin@example.com"}
            )
            mock_service.users().settings().updateAutoForwarding.assert_called_once()
            assert result["message"] == "Forwarding created"
            smtp.logger.info.assert_called()

    @pytest.mark.asyncio
    async def test_send_email(self, smtp: GmailSmtp) -> None:
        with (
            patch.object(smtp, "_get_gmail_service") as mock_get_service,
            patch("acb.adapters.smtp.gmail.MIMEMultipart") as mock_mime_multipart,
            patch("acb.adapters.smtp.gmail.MIMEText") as mock_mime_text,
            patch("acb.adapters.smtp.gmail.base64") as mock_base64,
        ):
            mock_service = MagicMock()
            mock_get_service.return_value = mock_service

            mock_message = MagicMock()
            mock_mime_multipart.return_value = mock_message

            mock_send = MagicMock()
            mock_service.users().messages().send.return_value = mock_send
            mock_send.execute.return_value = {"id": "message-id-123"}

            mock_base64.urlsafe_b64encode.return_value.decode.return_value = (
                "encoded-message"
            )

            result = await smtp.send_email(
                to="user@example.com",
                subject="Test Subject",
                body="Test Body",
            )

            mock_mime_multipart.assert_called_once_with("alternative")
            mock_message.__setitem__.assert_any_call("to", "user@example.com")
            mock_message.__setitem__.assert_any_call("subject", "Test Subject")
            mock_mime_text.assert_called_once_with("Test Body", "plain")
            mock_message.attach.assert_called_once_with(mock_mime_text.return_value)

            mock_service.users().messages().send.assert_called_once_with(
                userId="me", body={"raw": "encoded-message"}
            )

            assert result["id"] == "message-id-123"
            assert result["status"] == "sent"
            smtp.logger.info.assert_called_once()

    @pytest.mark.asyncio
    async def test_init(self, smtp: GmailSmtp) -> None:
        with (
            patch.object(smtp, "create_dns_records", AsyncMock()) as mock_create_dns,
            patch.object(smtp, "delete_routes", AsyncMock()) as mock_delete_routes,
            patch.object(smtp, "create_routes", AsyncMock()) as mock_create_routes,
        ):
            await smtp.init()

            mock_create_dns.assert_called_once()
            mock_delete_routes.assert_called_once()
            mock_create_routes.assert_called_once()



"""Simplified tests for the Memory Cache adapter."""

import pytest
from tests.test_interfaces import CacheTestInterface, MockCache


@pytest.fixture
async def cache() -> MockCache:
    cache = MockCache()
    await cache.init()
    return cache


@pytest.mark.unit
class TestMemoryCache(CacheTestInterface):
    pass



import hashlib
from pathlib import Path
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from anyio import Path as AsyncPath
from blake3 import blake3
from acb.actions.hash import hash


class TestHash:
    @pytest.mark.asyncio
    async def test_blake3_with_string(self) -> None:
        test_string = "test string"
        expected_hash = blake3(test_string.encode()).hexdigest()
        result = await hash.blake3(test_string)
        assert result == expected_hash

    @pytest.mark.asyncio
    async def test_blake3_with_bytes(self) -> None:
        test_bytes = b"test bytes"
        expected_hash = blake3(test_bytes).hexdigest()
        result = await hash.blake3(test_bytes)
        assert result == expected_hash

    @pytest.mark.asyncio
    async def test_blake3_with_file(self, tmp_path: Path) -> None:
        file_content = "test file content"
        file_path = AsyncPath(tmp_path) / "test_file.txt"

        with (
            patch.object(AsyncPath, "write_text", AsyncMock(return_value=None)),
            patch.object(
                AsyncPath, "read_bytes", AsyncMock(return_value=file_content.encode())
            ),
            patch.object(AsyncPath, "exists", AsyncMock(return_value=True)),
            patch.object(Path, "exists", MagicMock(return_value=True)),
        ):
            expected_hash = blake3(file_content.encode()).hexdigest()

            result_async = await hash.blake3(file_path)
            assert result_async == expected_hash

            result_path = await hash.blake3(file_path._path)
            assert result_path == expected_hash

    @pytest.mark.asyncio
    async def test_crc32c_with_string(self) -> None:
        test_string = "test string for crc32c"

        with patch("acb.actions.hash.crc32c") as mock_crc32c:
            mock_crc32c.return_value = 123456789

            result = await hash.crc32c(test_string)
            assert result == 123456789
            mock_crc32c.assert_called_with(test_string.encode())

    @pytest.mark.asyncio
    async def test_crc32c_with_bytes(self) -> None:
        test_bytes = b"test bytes for crc32c"

        with patch("acb.actions.hash.crc32c") as mock_crc32c:
            mock_crc32c.return_value = 987654321

            result = await hash.crc32c(test_bytes)
            assert result == 987654321
            mock_crc32c.assert_called_with(test_bytes)

    @pytest.mark.asyncio
    async def test_crc32c_with_file(self, tmp_path: Path) -> None:
        file_content = "test file content for crc32c"
        file_path = AsyncPath(tmp_path) / "test_crc32c.txt"

        with (
            patch.object(AsyncPath, "write_text", AsyncMock(return_value=None)),
            patch.object(AsyncPath, "read_text", AsyncMock(return_value=file_content)),
            patch.object(AsyncPath, "exists", AsyncMock(return_value=True)),
            patch.object(Path, "exists", MagicMock(return_value=True)),
            patch("acb.actions.hash.crc32c") as mock_crc32c,
        ):
            mock_crc32c.return_value = 987654321

            result_async = await hash.crc32c(file_path)
            assert result_async == 987654321
            mock_crc32c.assert_called_with(file_content.encode())

            result_path = await hash.crc32c(file_path._path)
            assert result_path == 987654321

    @pytest.mark.asyncio
    async def test_md5_with_string(self) -> None:
        test_string = "test string for md5"
        expected_hash = hashlib.md5(
            test_string.encode(), usedforsecurity=False
        ).hexdigest()
        result = await hash.md5(test_string)
        assert result == expected_hash

    @pytest.mark.asyncio
    async def test_md5_with_bytes(self) -> None:
        test_bytes = b"test bytes for md5"
        expected_hash = hashlib.md5(test_bytes, usedforsecurity=False).hexdigest()
        result = await hash.md5(test_bytes)
        assert result == expected_hash

    @pytest.mark.asyncio
    async def test_md5_with_file(self, tmp_path: Path) -> None:
        file_content = "test file content for md5"
        file_path = AsyncPath(tmp_path) / "test_md5.txt"

        with (
            patch.object(AsyncPath, "write_text", AsyncMock(return_value=None)),
            patch.object(AsyncPath, "read_text", AsyncMock(return_value=file_content)),
            patch.object(AsyncPath, "exists", AsyncMock(return_value=True)),
            patch.object(Path, "exists", MagicMock(return_value=True)),
        ):
            expected_hash = hashlib.md5(
                file_content.encode(), usedforsecurity=False
            ).hexdigest()

            result_async = await hash.md5(file_path)
            assert result_async == expected_hash

            result_path = await hash.md5(file_path._path)
            assert result_path == expected_hash



import typing as t
from contextlib import asynccontextmanager
from functools import cached_property

from google.cloud import firestore
from google.cloud.firestore_v1.base_query import FieldFilter
from acb.config import Config
from acb.depends import depends

from ._base import NosqlBase, NosqlBaseSettings


class NosqlSettings(NosqlBaseSettings):
    project_id: t.Optional[str] = None
    credentials_path: t.Optional[str] = None
    emulator_host: t.Optional[str] = None

    @depends.inject
    def __init__(self, config: Config = depends(), **values: t.Any) -> None:
        super().__init__(**values)
        if not self.project_id:
            self.project_id = config.app.project


class Nosql(NosqlBase):
    _transaction = None

    @cached_property
    def client(self) -> firestore.Client:
        if not self._client:
            kwargs = {}
            if self.config.nosql.project_id:
                kwargs["project"] = self.config.nosql.project_id
            if self.config.nosql.credentials_path:
                kwargs["credentials"] = self.config.nosql.credentials_path

            self._client = firestore.Client(**kwargs)
        return self._client

    @cached_property
    def db(self) -> firestore.Client:
        return self.client

    async def init(self) -> None:
        self.logger.info(
            f"Initializing Firestore connection for project {self.config.nosql.project_id}"
        )
        try:
            self.client.collection("test")
            self.logger.info("Firestore connection initialized successfully")
        except Exception as e:
            self.logger.error(f"Failed to initialize Firestore connection: {e}")
            raise

    def _get_collection_ref(self, collection: str) -> firestore.CollectionReference:
        prefix = self.config.nosql.collection_prefix
        return self.client.collection(f"{prefix}{collection}")

    def _convert_to_dict(self, doc: firestore.DocumentSnapshot) -> t.Dict[str, t.Any]:
        if not doc.exists:
            return {}

        data = doc.to_dict()
        if data is None:
            data = {}
        data["_id"] = doc.id
        return data

    def _prepare_document(self, document: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
        if "_id" in document:
            document = document.copy()
            del document["_id"]
        return document

    async def find(
        self, collection: str, filter: t.Dict[str, t.Any], **kwargs: t.Any
    ) -> t.List[t.Dict[str, t.Any]]:
        collection_ref = self._get_collection_ref(collection)
        query = collection_ref

        for key, value in filter.items():
            if key == "_id":
                continue
            query = query.where(filter=FieldFilter(key, "==", value))

        if "limit" in kwargs:
            query = query.limit(kwargs["limit"])

        if "order_by" in kwargs:
            for field in kwargs["order_by"]:
                direction = (
                    firestore.Query.DESCENDING
                    if field.startswith("-")
                    else firestore.Query.ASCENDING
                )
                field_name = field.removeprefix("-")
                query = query.order_by(field_name, direction=direction)

        docs = query.stream()

        results = [self._convert_to_dict(doc) for doc in docs]

        if "_id" in filter:
            results = [doc for doc in results if doc.get("_id") == filter["_id"]]

        return results

    async def find_one(
        self, collection: str, filter: t.Dict[str, t.Any], **kwargs: t.Any
    ) -> t.Optional[t.Dict[str, t.Any]]:
        if "_id" in filter:
            doc_ref = self._get_collection_ref(collection).document(filter["_id"])
            doc = doc_ref.get()
            if doc.exists:
                return self._convert_to_dict(doc)
            return None

        results = await self.find(collection, filter, limit=1, **kwargs)
        return results[0] if results else None

    async def insert_one(
        self, collection: str, document: t.Dict[str, t.Any], **kwargs: t.Any
    ) -> t.Any:
        collection_ref = self._get_collection_ref(collection)

        if "_id" in document:
            doc_id = document["_id"]
            doc_ref = collection_ref.document(doc_id)
            doc_ref.set(self._prepare_document(document))
        else:
            doc_ref = collection_ref.add(document)[1]
            doc_id = doc_ref.id

        return doc_id

    async def insert_many(
        self, collection: str, documents: t.List[t.Dict[str, t.Any]], **kwargs: t.Any
    ) -> t.List[t.Any]:
        ids = []
        batch = self.client.batch()
        collection_ref = self._get_collection_ref(collection)

        for doc in documents:
            if "_id" in doc:
                doc_id = doc["_id"]
                doc_ref = collection_ref.document(doc_id)
                batch.set(doc_ref, self._prepare_document(doc))
            else:
                doc_ref = collection_ref.document()
                batch.set(doc_ref, doc)
                doc_id = doc_ref.id

            ids.append(doc_id)

        batch.commit()
        return ids

    async def update_one(
        self,
        collection: str,
        filter: t.Dict[str, t.Any],
        update: t.Dict[str, t.Any],
        **kwargs: t.Any,
    ) -> t.Any:
        doc = await self.find_one(collection, filter)
        if not doc:
            return {"modified_count": 0}

        doc_id = doc["_id"]
        doc_ref = self._get_collection_ref(collection).document(doc_id)

        if "$set" in update:
            doc_ref.update(update["$set"])
        else:
            doc_ref.update(update)

        return {"modified_count": 1}

    async def update_many(
        self,
        collection: str,
        filter: t.Dict[str, t.Any],
        update: t.Dict[str, t.Any],
        **kwargs: t.Any,
    ) -> t.Any:
        docs = await self.find(collection, filter)
        if not docs:
            return {"modified_count": 0}

        batch = self.client.batch()
        collection_ref = self._get_collection_ref(collection)

        for doc in docs:
            doc_id = doc["_id"]
            doc_ref = collection_ref.document(doc_id)

            if "$set" in update:
                batch.update(doc_ref, update["$set"])
            else:
                batch.update(doc_ref, update)

        batch.commit()
        return {"modified_count": len(docs)}

    async def delete_one(
        self, collection: str, filter: t.Dict[str, t.Any], **kwargs: t.Any
    ) -> t.Any:
        doc = await self.find_one(collection, filter)
        if not doc:
            return {"deleted_count": 0}

        doc_id = doc["_id"]
        doc_ref = self._get_collection_ref(collection).document(doc_id)

        doc_ref.delete()

        return {"deleted_count": 1}

    async def delete_many(
        self, collection: str, filter: t.Dict[str, t.Any], **kwargs: t.Any
    ) -> t.Any:
        docs = await self.find(collection, filter)
        if not docs:
            return {"deleted_count": 0}

        batch = self.client.batch()
        collection_ref = self._get_collection_ref(collection)

        for doc in docs:
            doc_id = doc["_id"]
            doc_ref = collection_ref.document(doc_id)
            batch.delete(doc_ref)

        batch.commit()
        return {"deleted_count": len(docs)}

    async def count(
        self,
        collection: str,
        filter: t.Optional[t.Dict[str, t.Any]] = None,
        **kwargs: t.Any,
    ) -> int:
        docs = await self.find(collection, filter or {})
        return len(docs)

    async def aggregate(
        self, collection: str, pipeline: t.List[t.Dict[str, t.Any]], **kwargs: t.Any
    ) -> t.List[t.Dict[str, t.Any]]:
        docs = await self.find(collection, {})

        for stage in pipeline:
            if "$match" in stage:
                filter_dict = stage["$match"]
                docs = [
                    doc
                    for doc in docs
                    if all(doc.get(k) == v for k, v in filter_dict.items())
                ]
            elif "$project" in stage:
                projection = stage["$project"]
                docs = [{k: doc[k] for k in projection if k in doc} for doc in docs]
            elif "$limit" in stage:
                docs = docs[: stage["$limit"]]
            elif "$skip" in stage:
                docs = docs[stage["$skip"] :]

        return docs

    @asynccontextmanager
    async def transaction(self) -> t.AsyncGenerator[None, None]:
        transaction = self.client.transaction()
        try:
            self._transaction = transaction
            with transaction:
                yield
        except Exception as e:
            self.logger.error(f"Transaction failed: {e}")
            raise
        finally:
            self._transaction = None


depends.set(Nosql)



import typing as t
from functools import cached_property
from typing import Any, cast
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from anyio import Path as AsyncPath
from acb.config import Config

from ._base import (
    StorageBase,
    StorageBaseSettings,
    StorageBucket,
    StorageFile,
    StorageImage,
)
from .file import Storage as FileStorage
from .memory import Storage as MemoryStorage


@pytest.fixture
def mock_storage_settings() -> MagicMock:
    settings = MagicMock()
    settings.buckets = {
        "test": "test-bucket",
        "media": "media-bucket",
        "templates": "templates-bucket",
    }
    settings.prefix = "test-prefix"
    mock_path = MagicMock(spec=AsyncPath)
    mock_path.__truediv__.return_value = mock_path
    settings.local_path = mock_path
    settings.local_fs = True
    settings.memory_fs = False
    settings.cors = None
    return settings


@pytest.fixture
def mock_app_settings() -> MagicMock:
    settings = MagicMock()
    settings.name = "acb"
    return settings


@pytest.fixture
def mock_config(
    mock_storage_settings: MagicMock, mock_app_settings: MagicMock
) -> MagicMock:
    config = MagicMock(spec=Config)
    type(config).__getattr__ = MagicMock(return_value=mock_storage_settings)
    config.__dict__["storage"] = mock_storage_settings
    config.app = mock_app_settings
    config.deployed = False
    return config


@pytest.fixture
def mock_client() -> AsyncMock:
    client = AsyncMock()
    client._pipe_file = AsyncMock()
    client._rm_file = AsyncMock()
    client._exists = AsyncMock(return_value=True)
    client._info = AsyncMock(
        return_value={"size": 100, "timeCreated": "2023-01-01", "updated": "2023-01-02"}
    )
    client._ls = AsyncMock(return_value=["file1.txt", "file2.txt"])
    client._mkdir = AsyncMock()
    client._sign = AsyncMock(return_value="https://signed-url.example.com")
    client.url = MagicMock(return_value="https://example.com/file.txt")

    async_cm = AsyncMock()
    async_cm.__aenter__.return_value.read = AsyncMock(return_value=b"test content")
    client.open = MagicMock(return_value=async_cm)

    client.pipe_file = MagicMock()
    client.info = MagicMock(
        return_value={"name": "test.txt", "size": 100, "type": "file"}
    )
    client.modified = MagicMock()
    client.created = MagicMock()
    client.isfile = MagicMock(return_value=True)
    client.set_cors = MagicMock()
    return client


@pytest.fixture
def mock_adapter() -> MagicMock:
    adapter = MagicMock()
    adapter.name = "memory"
    return adapter


@pytest.fixture
def mock_storage_bucket(
    mock_client: AsyncMock,
    mock_config: MagicMock,
) -> StorageBucket:
    with patch("acb.depends.depends.get", return_value=mock_config):

        class TestStorageBucket(StorageBucket):
            _config_value: Config | None = None

            @property
            def config(self) -> Config:  # type: ignore
                if self._config_value is None:
                    self._config_value = mock_config
                return self._config_value

            def get_path(self, path: AsyncPath | str) -> str:
                return f"test-bucket/{path}"

        bucket = TestStorageBucket(mock_client, "test")
        bucket.bucket = "test-bucket"
        return bucket


@pytest.fixture
def mock_storage_base(
    mock_client: AsyncMock, mock_config: MagicMock, mock_storage_bucket: StorageBucket
) -> StorageBase:
    with patch("acb.depends.depends.get", return_value=mock_config):

        class TestStorageBase(StorageBase):
            _config_value: Config | None = None

            @property
            def config(self) -> Config:  # type: ignore
                if self._config_value is None:
                    self._config_value = mock_config
                return self._config_value

            @cached_property
            def client(self) -> Any:
                return mock_client

            async def init(self) -> None:
                self.test = mock_storage_bucket
                self.media = mock_storage_bucket
                self.templates = mock_storage_bucket

        storage = TestStorageBase()
        storage.test = mock_storage_bucket
        storage.media = mock_storage_bucket
        storage.templates = mock_storage_bucket
        return storage


class TestStorageBaseSettings:
    def test_init(
        self,
        mock_config: MagicMock,
        mock_storage_settings: MagicMock,
    ) -> None:
        settings = MagicMock(spec=StorageBaseSettings)
        settings.prefix = "test_prefix"
        settings.buckets = {
            "test": "test-bucket",
            "media": "media-bucket",
            "templates": "templates-bucket",
        }
        settings.local_fs = True
        settings.memory_fs = False

        assert settings.prefix is not None
        assert "test" in settings.buckets
        assert "media" in settings.buckets
        assert "templates" in settings.buckets


class TestStorageBucket:
    @pytest.mark.anyio
    async def test_get_path(
        self,
        mock_storage_bucket: StorageBucket,
    ) -> None:
        path = AsyncPath("test.txt")
        result = mock_storage_bucket.get_path(path)
        assert isinstance(result, str)
        assert "test-bucket" in result

    @pytest.mark.anyio
    async def test_get_url(
        self, mock_storage_bucket: StorageBucket, mock_client: AsyncMock
    ) -> None:
        path = AsyncPath("test.txt")
        url = mock_storage_bucket.get_url(path)
        assert url == "https://example.com/file.txt"
        mock_client.url.assert_called_once()

    @pytest.mark.anyio
    async def test_get_signed_url(
        self, mock_storage_bucket: StorageBucket, mock_client: AsyncMock
    ) -> None:
        path = AsyncPath("test.txt")
        url = await mock_storage_bucket.get_signed_url(path)
        assert url == "https://signed-url.example.com"
        mock_client._sign.assert_called_once()

    @pytest.mark.anyio
    async def test_stat(
        self, mock_storage_bucket: StorageBucket, mock_client: AsyncMock
    ) -> None:
        path = AsyncPath("test.txt")
        stat = await mock_storage_bucket.stat(path)
        assert stat["size"] == 100
        assert "timeCreated" in stat
        assert "updated" in stat
        mock_client._info.assert_called_once()

    @pytest.mark.anyio
    async def test_list(
        self, mock_storage_bucket: StorageBucket, mock_client: AsyncMock
    ) -> None:
        path = AsyncPath("directory/")
        files = await mock_storage_bucket.list(path)
        assert len(files) == 2
        assert "file1.txt" in files
        assert "file2.txt" in files
        mock_client._ls.assert_called_once()

    @pytest.mark.anyio
    async def test_exists(
        self, mock_storage_bucket: StorageBucket, mock_client: AsyncMock
    ) -> None:
        async def mock_exists(path: AsyncPath) -> bool:
            await mock_client._exists(mock_storage_bucket.get_path(path))
            mock_client._exists.assert_called_once()
            return True

        mock_storage_bucket.exists = mock_exists
        path = AsyncPath("test.txt")
        exists = await mock_storage_bucket.exists(path)
        assert exists

    @pytest.mark.anyio
    async def test_create_bucket(
        self, mock_storage_bucket: StorageBucket, mock_client: AsyncMock
    ) -> None:
        path = AsyncPath("new-bucket/")
        await mock_storage_bucket.create_bucket(path)
        mock_client._mkdir.assert_called_once()

    @pytest.mark.anyio
    async def test_open(
        self, mock_storage_bucket: StorageBucket, mock_client: AsyncMock
    ) -> None:
        async def mock_open(path: AsyncPath) -> t.BinaryIO:
            mock_client.open(mock_storage_bucket.get_path(path))
            mock_client.open.assert_called_once()
            return t.cast(t.BinaryIO, b"test content")

        mock_storage_bucket.open = mock_open
        path = AsyncPath("test.txt")
        content = await mock_storage_bucket.open(path)
        assert content == b"test content"

    @pytest.mark.anyio
    async def test_open_file_not_found(
        self, mock_storage_bucket: StorageBucket, mock_client: AsyncMock
    ) -> None:
        async def mock_open_not_found(path: AsyncPath) -> t.BinaryIO:
            raise FileNotFoundError(f"File not found: {path}")

        mock_storage_bucket.open = mock_open_not_found
        path = AsyncPath("nonexistent.txt")
        with pytest.raises(FileNotFoundError):
            await mock_storage_bucket.open(path)

    @pytest.mark.anyio
    async def test_write(
        self, mock_storage_bucket: StorageBucket, mock_client: AsyncMock
    ) -> None:
        async def mock_write(path: AsyncPath, data: bytes) -> None:
            await mock_client._pipe_file(mock_storage_bucket.get_path(path), data)
            mock_client._pipe_file.assert_called_once()
            return None

        mock_storage_bucket.write = mock_write
        path = AsyncPath("test.txt")
        data = b"test content"
        await mock_storage_bucket.write(path, data)

    @pytest.mark.anyio
    async def test_delete(
        self, mock_storage_bucket: StorageBucket, mock_client: AsyncMock
    ) -> None:
        path = AsyncPath("test.txt")
        await mock_storage_bucket.delete(path)
        mock_client._rm_file.assert_called_once()


class TestStorageBase:
    def test_client_property(self, mock_storage_base: StorageBase) -> None:
        mock_client = MagicMock()
        mock_storage_base.client = mock_client

        client = mock_storage_base.client

        assert client is mock_client


class TestStorageFile:
    def test_init(self, mock_storage_bucket: StorageBucket) -> None:
        file = StorageFile(name="test.txt", storage=mock_storage_bucket)
        assert file._storage == mock_storage_bucket
        assert file._name == "test.txt"

    def test_name_property(self, mock_storage_bucket: StorageBucket) -> None:
        file = StorageFile(name="test.txt", storage=mock_storage_bucket)
        assert file.name == "test.txt"

    def test_path_property(
        self,
        mock_storage_bucket: StorageBucket,
    ) -> None:
        mock_storage_bucket.get_path = lambda path: f"test-bucket/{path}"  # type: ignore[assignment]
        file = StorageFile(name="test.txt", storage=mock_storage_bucket)
        assert isinstance(file.path, str)
        assert "test-bucket" in file.path

    @pytest.mark.anyio
    async def test_size_property(
        self, mock_storage_bucket: StorageBucket, mock_client: AsyncMock
    ) -> None:
        mock_client._info.return_value = {"size": 100}
        file = StorageFile(name="test.txt", storage=mock_storage_bucket)
        size = await file.size
        assert size == 100

    @pytest.mark.anyio
    async def test_open_method(
        self, mock_storage_bucket: StorageBucket, mock_client: AsyncMock
    ) -> None:
        async def mock_open(path: AsyncPath) -> t.BinaryIO:
            return t.cast(t.BinaryIO, b"test content")

        mock_storage_bucket.open = mock_open
        file = StorageFile(name="test.txt", storage=mock_storage_bucket)
        content = await file.open()
        assert content == b"test content"

    @pytest.mark.anyio
    async def test_write_method(
        self, mock_storage_bucket: StorageBucket, mock_client: AsyncMock
    ) -> None:
        write_called: bool = False

        async def mock_write(path: AsyncPath, data: bytes) -> None:
            nonlocal write_called
            write_called = True
            return None

        mock_storage_bucket.write = mock_write
        file = StorageFile(name="test.txt", storage=mock_storage_bucket)
        data = b"test content"
        await file.write(cast(t.BinaryIO, data))
        assert write_called


class TestStorageImage:
    def test_init(self, mock_storage_bucket: StorageBucket) -> None:
        image = StorageImage(
            name="image.jpg", storage=mock_storage_bucket, height=100, width=200
        )
        assert image._storage == mock_storage_bucket
        assert image._name == "image.jpg"
        assert image._height == 100
        assert image._width == 200

    def test_height_property(self, mock_storage_bucket: StorageBucket) -> None:
        image = StorageImage(
            name="image.jpg", storage=mock_storage_bucket, height=100, width=200
        )
        assert image.height == 100

    def test_width_property(self, mock_storage_bucket: StorageBucket) -> None:
        image = StorageImage(
            name="image.jpg", storage=mock_storage_bucket, height=100, width=200
        )
        assert image.width == 200


class TestFileStorage:
    def test_init(self) -> None:
        with patch("acb.adapters.storage.file.DirFileSystem") as mock_dir_fs:
            mock_dir_fs.__name__ = "DirFileSystem"

            storage = FileStorage()

            storage.file_system = mock_dir_fs

            assert storage.file_system.__name__ == "DirFileSystem"


class TestMemoryStorage:
    def test_init(self) -> None:
        with patch("acb.adapters.storage.memory.MemoryFileSystem") as mock_memory_fs:
            storage = MemoryStorage()

            storage.file_system = mock_memory_fs

            assert storage.file_system == mock_memory_fs


@pytest.mark.anyio
async def test_integration_memory_storage(
    mock_config: MagicMock,
) -> None:
    with patch("acb.depends.depends.get", return_value=mock_config):

        class SimpleMemoryStorage:
            def __init__(self) -> None:
                self.files: dict[str, bytes] = {}
                self.test = self

            async def write(self, path: AsyncPath, data: bytes) -> None:
                self.files[str(path)] = data

            async def exists(self, path: AsyncPath) -> bool:
                return str(path) in self.files

            async def open(self, path: AsyncPath) -> bytes:
                if str(path) not in self.files:
                    raise FileNotFoundError(f"File not found: {path}")
                return self.files[str(path)]

            async def stat(self, path: AsyncPath) -> dict[str, int]:
                if str(path) not in self.files:
                    raise FileNotFoundError(f"File not found: {path}")
                return {"size": len(self.files[str(path)])}

            async def list(self, path: AsyncPath) -> list[str]:
                return list(self.files.keys())

            async def delete(self, path: AsyncPath) -> None:
                if str(path) in self.files:
                    del self.files[str(path)]

        storage = SimpleMemoryStorage()

        test_data: bytes = b"Hello, World!"
        test_path: AsyncPath = AsyncPath("test.txt")

        await storage.write(test_path, test_data)
        assert await storage.exists(test_path)

        content = await storage.open(test_path)
        assert content == test_data

        stat = await storage.stat(test_path)
        assert stat["size"] == len(test_data)

        files = await storage.list(AsyncPath("/"))
        assert "test.txt" in files

        await storage.delete(test_path)
        with pytest.raises(FileNotFoundError):
            await storage.open(test_path)
